"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.addTable = void 0;
const path_1 = __importDefault(require("path"));
const client_1 = require("../../api/client");
const msg_1 = __importDefault(require("../../utils/msg"));
const logger_1 = require("../../logger");
const file_1 = require("../../utils/file");
const global_1 = require("../../config/global");
const constants_1 = __importDefault(require("../../constants"));
const migrations_1 = require("../../config/migrations");
const formatters_1 = require("../../utils/formatters");
const pm_1 = require("@spec.dev/pm");
const migrate_1 = require("../migrate");
const migrations_2 = require("../../config/migrations");
const CMD = 'table';
function addTableCmd(cmd) {
    cmd.command(CMD)
        .description('Add a new Live Table to your current project')
        .argument('source', 'Live Table data source')
        .option('--name <type>', 'Table name')
        .option('--migrate', 'Whether to also immediately run the new table migration')
        .option('--config-only', 'Avoid generating the SQL migration')
        .action(addTable);
}
/**
 * Add the SQL migration and Spec config for a
 * new live table to your current project.
 */
function addTable(source, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        // Get the current project id.
        const { data: projectId, error } = (0, global_1.getCurrentProjectId)();
        if (error) {
            (0, logger_1.logFailure)(error);
            return;
        }
        if (!projectId) {
            (0, logger_1.logWarning)(msg_1.default.CHOOSE_PROJECT_MESSAGE);
            return;
        }
        // Get the location of the project (/path/to/project)
        const { data: projectInfo, error: getProjectError } = (0, global_1.getProjectInfo)(projectId);
        if (getProjectError) {
            (0, logger_1.logFailure)(error);
            return;
        }
        const { location: projectDirPath } = projectInfo || {};
        if (!projectDirPath) {
            (0, logger_1.logWarning)(msg_1.default.NO_PROJECT_LOCATION);
            return;
        }
        const specConfigDir = path_1.default.join(projectDirPath, constants_1.default.SPEC_CONFIG_DIR_NAME);
        if (!(0, file_1.fileExists)(specConfigDir)) {
            (0, logger_1.logWarning)(`Spec config directory doesn't exist at ${specConfigDir}.`);
            return;
        }
        const projectConfigPath = path_1.default.join(specConfigDir, constants_1.default.PROJECT_CONFIG_FILE_NAME);
        if (!(0, file_1.fileExists)(projectConfigPath)) {
            (0, logger_1.logWarning)(`project.toml is missing inside the current project. Try running "spec init" and then try again.`);
            return;
        }
        // Get live object version for given source.
        const { lov, error: lovError } = yield client_1.client.getLiveObjectVersion(source);
        if (lovError) {
            lovError.toLowerCase().includes('not found')
                ? (0, logger_1.logWarning)(`No live object version found for "${source}".`)
                : (0, logger_1.logFailure)(`Error resolving live object version: ${lovError}`);
            return;
        }
        const { nsp, name } = (0, formatters_1.fromNamespacedVersion)(lov.name);
        const isContractEvent = nsp.includes('.');
        // Add "_event" as a suffix to the table if it's an event
        // table and the table name wasn't explicitly specified.
        let tableName = opts.name || (0, formatters_1.camelToSnake)(name);
        if (!opts.name && isContractEvent) {
            tableName += '_event';
        }
        const tablePath = [migrations_2.schemaName, tableName].join('.');
        if (!opts.configOnly) {
            // Create new table migration from this live object version's structure.
            const { up, down, migration } = (0, migrations_1.generateCreateTableMigrationFromLov)(lov, tableName);
            const { error: migrationError } = (0, migrations_1.saveMigration)(projectDirPath, migration, up, down);
            if (migrationError) {
                (0, logger_1.logFailure)(migrationError);
                return;
            }
            if (opts.migrate && !(yield (0, migrate_1.migrate)({}, false)))
                return;
        }
        // Add the default live table config to project.toml
        const liveColumns = {};
        for (const { name } of lov.properties) {
            liveColumns[(0, formatters_1.camelToSnake)(name)] = { property: name };
        }
        try {
            (0, pm_1.upsertLiveColumns)(projectDirPath, {
                tablePath,
                liveObjectVersionId: lov.name,
                liveColumns,
                filters: [],
                uniqueBy: lov.uniqueBy,
            });
        }
        catch (err) {
            (0, logger_1.logFailure)(err);
            return;
        }
        (0, logger_1.logSuccess)(`Added new live table "${tablePath.split('.')[1]}".`);
    });
}
exports.addTable = addTable;
exports.default = addTableCmd;
//# sourceMappingURL=table.js.map