"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.migrate = void 0;
const logger_1 = require("../logger");
const msg_1 = __importDefault(require("../utils/msg"));
const global_1 = require("../config/global");
const connect_1 = require("../config/connect");
const constants_1 = __importDefault(require("../constants"));
const path_1 = __importDefault(require("path"));
const db_1 = require("../db");
const file_1 = require("../utils/file");
const migrations_1 = require("../config/migrations");
const formatters_1 = require("../utils/formatters");
const CMD = 'migrate';
function addMigrateCmd(program) {
    program
        .command(CMD)
        .description("Sync your current project's SQL migrations to a database")
        .option('--env <type>', 'Project environment to run migrations against')
        .action(migrate);
}
/**
 * Run the SQL migrations for your current project (inside .spec/migrations)
 * against one of your database environments listed in connect.toml.
 */
function migrate(options, logWhenNoAction = true) {
    return __awaiter(this, void 0, void 0, function* () {
        // Get the current project id.
        const { data: projectId, error } = (0, global_1.getCurrentProjectId)();
        if (error) {
            (0, logger_1.logFailure)(error);
            return false;
        }
        if (!projectId) {
            (0, logger_1.logWarning)(msg_1.default.CHOOSE_PROJECT_MESSAGE);
            return false;
        }
        // Get the current project env (local, staging, prod, etc).
        let env = options.env;
        if (!env) {
            const { data, error: getEnvError } = (0, global_1.getCurrentProjectEnv)();
            env = data;
            if (getEnvError) {
                (0, logger_1.logFailure)(error);
                return false;
            }
            if (!env) {
                (0, logger_1.logWarning)(msg_1.default.NO_CURRENT_ENV_MESSAGE);
                return false;
            }
        }
        // Get the location of the project (/path/to/project)
        const { data: projectInfo, error: getProjectError } = (0, global_1.getProjectInfo)(projectId);
        if (getProjectError) {
            (0, logger_1.logFailure)(error);
            return false;
        }
        const { location: projectDirPath } = projectInfo || {};
        if (!projectDirPath) {
            (0, logger_1.logWarning)(msg_1.default.NO_PROJECT_LOCATION);
            return false;
        }
        const specConfigDir = path_1.default.join(projectDirPath, constants_1.default.SPEC_CONFIG_DIR_NAME);
        if (!(0, file_1.fileExists)(specConfigDir)) {
            (0, logger_1.logWarning)(`Spec config directory doesn't exist at ${specConfigDir}.`);
            return false;
        }
        const migrationsDir = path_1.default.join(specConfigDir, constants_1.default.MIGRATIONS_DIR_NAME);
        if (!(0, file_1.fileExists)(migrationsDir)) {
            logWhenNoAction && (0, logger_1.log)(`Migrations up-to-date.`);
            return true;
        }
        // Get DB info from connection config file.
        let { data: dbConfig, error: getDBConfigError } = (0, connect_1.getDBConfig)(projectDirPath, env);
        if (getDBConfigError) {
            (0, logger_1.logFailure)(getDBConfigError);
            return false;
        }
        if (!dbConfig) {
            (0, logger_1.logWarning)(`No environment named "${env}" inside ${constants_1.default.CONNECTION_CONFIG_FILE_NAME}.`);
            return false;
        }
        if (!dbConfig.name) {
            (0, logger_1.logWarning)(`No database "name" specified for the "${env}" environment in ${constants_1.default.CONNECTION_CONFIG_FILE_NAME}.`);
            return false;
        }
        const { data: connParams, error: formatError } = (0, db_1.resolveDBConnectionParams)(dbConfig);
        if (formatError) {
            (0, logger_1.logWarning)(`Invalid database environment "${env}" — ${formatError}`);
            return false;
        }
        // Initialize database for usage with Spec (if not already).
        const url = (0, formatters_1.asPostgresUrl)(connParams);
        const { error: initDBError } = yield (0, db_1.initDatabase)(connParams.name);
        if (initDBError) {
            (0, logger_1.logFailure)(`Error initializing database for Spec: ${initDBError}`);
            return false;
        }
        // Apply any new migrations.
        return yield (0, migrations_1.syncMigrations)(migrationsDir, url, env, logWhenNoAction);
    });
}
exports.migrate = migrate;
exports.default = addMigrateCmd;
//# sourceMappingURL=migrate.js.map