"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const logger_1 = require("../logger");
const msg_1 = __importDefault(require("../utils/msg"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const time_1 = require("../utils/time");
const file_1 = require("../utils/file");
const client_1 = require("../api/client");
const formatters_1 = require("../utils/formatters");
const auth_1 = require("../utils/auth");
const ora_1 = __importDefault(require("ora"));
const types_1 = require("../types");
const chalk_1 = __importDefault(require("chalk"));
const ansi_diff_stream_1 = __importDefault(require("ansi-diff-stream"));
const progress_string_1 = __importDefault(require("progress-string"));
const open_1 = __importDefault(require("open"));
const date_1 = require("../utils/date");
const POLL_INTERVAL = 1000;
const CMD = 'publish';
function addPublishCmd(program) {
    program
        .command(CMD)
        .description('Publish a Live Table')
        .argument('folder', 'Folder of the Live Table')
        .option('--open', 'Open the Live Table in the Spec ecosytem once published')
        .action(publish);
}
/**
 * Publish a Live Table.
 */
function publish(givenFolderPath, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const { token: sessionToken, error: sessionTokenError } = (0, auth_1.getSessionToken)();
        if (sessionTokenError) {
            (0, logger_1.logFailure)(sessionTokenError);
            return;
        }
        if (!sessionToken) {
            (0, logger_1.log)(msg_1.default.AUTH_REQUIRED_MESSAGE);
            return;
        }
        const folderPath = path_1.default.join(process.cwd(), givenFolderPath);
        const manifestPath = path_1.default.join(folderPath, 'manifest.json');
        if (!(0, file_1.fileExists)(manifestPath)) {
            (0, logger_1.logWarning)(`No Live Table found within ${folderPath}`);
            return;
        }
        const manifest = JSON.parse(fs_1.default.readFileSync(manifestPath, 'utf8'));
        const { namespace, name, version } = manifest;
        const { uid, error } = yield client_1.client.publishLiveObjectVersion(namespace, name, version, givenFolderPath, sessionToken);
        if (error) {
            (0, logger_1.logFailure)(`Error publishing Live Table: ${error}`);
            return;
        }
        yield pollForPublishResult(namespace, uid, sessionToken, manifest, opts);
    });
}
function pollForPublishResult(nsp, uid, sessionToken, manifest, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const namespacedVersion = (0, formatters_1.toNamespacedVersion)(manifest.namespace, manifest.name, manifest.version);
        const progressBar = (0, progress_string_1.default)({
            width: 40,
            total: 1,
            incomplete: ' ',
            complete: '.',
        });
        const spinnerText = `Publishing ${namespacedVersion}`;
        const diff = (0, ansi_diff_stream_1.default)();
        const logProgress = (value, cursor, done) => {
            const lines = [];
            const date = (0, date_1.formatDate)(cursor);
            lines.push(chalk_1.default.gray(chalk_1.default.dim('...' + ' '.repeat(spinnerText.length))));
            const suffix = value === 1 ? chalk_1.default.cyanBright(date) : date;
            lines.push(`${namespacedVersion} ${chalk_1.default.cyan(progressBar(value))} ${suffix}`);
            lines.push(chalk_1.default.gray(chalk_1.default.dim('...')));
            diff.write(lines.join('\n'));
        };
        diff.pipe(process.stdout);
        const spinner = (0, ora_1.default)({
            text: spinnerText,
            stream: process.stdout,
        }).start();
        // Poll for progress of the publishing live object job.
        while (true) {
            const { status, cursor, metadata, failed, error } = yield client_1.client.getPublishLiveObjectVersionJob(sessionToken, uid);
            if (error && error.toLowerCase().includes('parsing json'))
                continue;
            // Job failed.
            if (failed || error) {
                spinner.stop();
                (0, logger_1.logFailure)(error || 'Publishing Live Table failed.');
                return;
            }
            // Indexing.
            if (status === types_1.PublishLiveObjectVersionJobStatus.Indexing) {
                if (metadata.startCursor) {
                    const start = new Date(metadata.startCursor).getTime();
                    const currentCursor = new Date(cursor || start);
                    const current = currentCursor.getTime();
                    const end = new Date().getTime();
                    const total = end - start;
                    const progress = current - start;
                    const fraction = progress / total;
                    logProgress(fraction, currentCursor);
                }
                spinner.text = `Indexing Live Table`;
            }
            // Done.
            if (status === types_1.PublishLiveObjectVersionJobStatus.Complete) {
                logProgress(1, new Date(cursor));
                spinner.stop();
                (0, logger_1.logSuccess)(`Successfully published ${namespacedVersion}`);
                let liveTableUrl;
                if (metadata.liveObjectUid) {
                    liveTableUrl = (0, formatters_1.toLiveTableUrl)(nsp, metadata.liveObjectUid);
                    (0, logger_1.log)(chalk_1.default.dim('View live:'));
                    (0, logger_1.log)(liveTableUrl);
                }
                if (opts.open && liveTableUrl) {
                    yield (0, open_1.default)(liveTableUrl);
                }
                return;
            }
            yield (0, time_1.sleep)(POLL_INTERVAL);
        }
    });
}
exports.default = addPublishCmd;
//# sourceMappingURL=publish.js.map