"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.start = void 0;
const path_1 = __importDefault(require("path"));
const logger_1 = require("../logger");
const connect_1 = require("../config/connect");
const global_1 = require("../config/global");
const db_1 = require("../db");
const client_1 = require("../utils/client");
const msg_1 = __importDefault(require("../utils/msg"));
const constants_1 = __importDefault(require("../constants"));
const migrate_1 = require("./migrate");
const file_1 = require("../utils/file");
const CMD = 'start';
function addStartCmd(program) {
    program.command(CMD).description('Run the Spec client locally').action(start);
}
/**
 * Start Spec locally, connecting to your local Postgres database.
 */
function start() {
    return __awaiter(this, void 0, void 0, function* () {
        // Get the current project id.
        const { data: projectId, error } = (0, global_1.getCurrentProjectId)();
        if (error) {
            (0, logger_1.logFailure)(error);
            return;
        }
        if (!projectId) {
            (0, logger_1.logWarning)(msg_1.default.CHOOSE_PROJECT_MESSAGE);
            return;
        }
        // Get the current project env (local, staging, prod, etc).
        const { data: projectEnv, error: getEnvError } = (0, global_1.getCurrentProjectEnv)();
        if (getEnvError) {
            (0, logger_1.logFailure)(error);
            return;
        }
        if (!projectEnv) {
            (0, logger_1.logWarning)(msg_1.default.NO_CURRENT_ENV_MESSAGE);
            return;
        }
        // Get the location of the project (/path/to/project)
        const { data: projectInfo, error: getProjectError } = (0, global_1.getProjectInfo)(projectId);
        if (getProjectError) {
            (0, logger_1.logFailure)(error);
            return;
        }
        const { location: projectDirPath } = projectInfo || {};
        if (!projectDirPath) {
            (0, logger_1.logWarning)(msg_1.default.NO_PROJECT_LOCATION);
            return;
        }
        const specConfigDir = path_1.default.join(projectDirPath, constants_1.default.SPEC_CONFIG_DIR_NAME);
        if (!(0, file_1.fileExists)(specConfigDir)) {
            (0, logger_1.logWarning)(`Spec config directory doesn't exist at ${specConfigDir}.`);
            return;
        }
        // Get the current project credentials from the global creds file.
        const { data: creds, error: credsError } = (0, global_1.getProjectCreds)(projectId);
        if (credsError) {
            (0, logger_1.logFailure)(`Error finding project credentials: ${credsError}`);
            return;
        }
        if (!(creds === null || creds === void 0 ? void 0 : creds.apiKey)) {
            (0, logger_1.logWarning)(msg_1.default.CHOOSE_PROJECT_MESSAGE);
            return;
        }
        // Get DB info from connection config file.
        let { data: dbConfig, error: getDBConfigError } = (0, connect_1.getDBConfig)(projectDirPath, projectEnv);
        if (getDBConfigError) {
            (0, logger_1.logFailure)(getDBConfigError);
            return;
        }
        if (!dbConfig) {
            (0, logger_1.logWarning)(`No environment named "${projectEnv}" inside ${constants_1.default.CONNECTION_CONFIG_FILE_NAME}.`);
            return;
        }
        if (!dbConfig.name) {
            (0, logger_1.logWarning)(`No database "name" specified for the "${projectEnv}" environment in ${constants_1.default.CONNECTION_CONFIG_FILE_NAME}.`);
            return;
        }
        // Ensure psql is installed.
        if (!(0, db_1.psqlInstalled)()) {
            (0, logger_1.logWarning)(msg_1.default.INSTALL_PSQL);
            return;
        }
        // Ensure spec-client is installed.
        if (!(0, client_1.specClientInstalled)()) {
            (0, logger_1.logWarning)(msg_1.default.INSTALL_SPEC_CLIENT);
            return;
        }
        // Initialize database for usage with Spec.
        const { newlyAssignedPassword, error: initDBError } = yield (0, db_1.initDatabase)(dbConfig.name);
        if (initDBError) {
            (0, logger_1.logFailure)(`Error initializing database: ${initDBError}`);
            return;
        }
        if (newlyAssignedPassword) {
            dbConfig.user = constants_1.default.SPEC_DB_USER;
            dbConfig.password = newlyAssignedPassword;
        }
        // Run any new table migrations.
        if (!(yield (0, migrate_1.migrate)({}, false)))
            return;
        // Run the Spec client.
        const { error: startError } = yield (0, client_1.startSpec)(projectId, specConfigDir, creds.apiKey, dbConfig);
        if (startError) {
            (0, logger_1.logFailure)(`Error starting Spec: ${startError}`);
        }
    });
}
exports.start = start;
exports.default = addStartCmd;
//# sourceMappingURL=start.js.map