"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.syncContracts = void 0;
const client_1 = require("../../api/client");
const msg_1 = __importDefault(require("../../utils/msg"));
const logger_1 = require("../../logger");
const types_1 = require("../../types");
const auth_1 = require("../../utils/auth");
const chains_1 = require("../../utils/chains");
const time_1 = require("../../utils/time");
const abi_1 = require("../../utils/abi");
const constants_1 = __importDefault(require("../../constants"));
const validators_1 = require("../../utils/validators");
const progress_string_1 = __importDefault(require("progress-string"));
const ansi_diff_stream_1 = __importDefault(require("ansi-diff-stream"));
const ora_1 = __importDefault(require("ora"));
const chalk_1 = __importDefault(require("chalk"));
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
const process_1 = __importDefault(require("process"));
const file_1 = require("../../utils/file");
const open_1 = __importDefault(require("open"));
const formatters_1 = require("../../utils/formatters");
const CMD = 'contracts';
const POLL_INTERVAL = 1000;
function addSyncContractsCmd(cmd) {
    cmd.command(CMD)
        .description('Sync contract groups from a JSON file')
        .option('-f, --file <type>', 'Contract groups file path')
        .option('--open', 'View your contract groups in the Spec ecosytem once synced')
        .action(syncContracts);
}
function syncContracts(opts) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        // Get authed user's session token (if any).
        const { token: sessionToken, error } = (0, auth_1.getSessionToken)();
        if (error) {
            (0, logger_1.logFailure)(error);
            return;
        }
        if (!sessionToken) {
            (0, logger_1.log)(msg_1.default.AUTH_REQUIRED_MESSAGE);
            return;
        }
        // Read/resolve contract groups from file.
        const { contents, error: fileError } = readContractGroupsSpecFile(opts.file);
        if (fileError) {
            (0, logger_1.logWarning)(fileError);
            return;
        }
        // Validate file contents and format into payload.
        const { payload, error: payloadError } = resolveContractGroupsPayload(contents);
        if (payloadError || !payload) {
            payloadError && (0, logger_1.logWarning)(payloadError);
            return;
        }
        // Sync contracts to Spec.
        const resp = yield client_1.client.registerContracts(sessionToken, payload);
        if (resp.error) {
            (0, logger_1.logFailure)(`Error syncing contracts: ${resp.error}`);
            return;
        }
        if (!((_a = resp.groups) === null || _a === void 0 ? void 0 : _a.length)) {
            (0, logger_1.log)(`No changes to sync.`);
            return;
        }
        yield pollForRegistrationResult(resp, sessionToken, opts.open);
    });
}
exports.syncContracts = syncContracts;
function readContractGroupsSpecFile(filePath) {
    const resolvedPath = path_1.default.resolve(filePath || constants_1.default.DEFAULT_CONTRACT_GROUPS_SPEC_FILE_NAME);
    if (!(0, file_1.fileExists)(resolvedPath)) {
        return { error: `No file found at ${resolvedPath}` };
    }
    try {
        const contents = JSON.parse(fs_1.default.readFileSync(resolvedPath, 'utf-8')) || {};
        return { contents };
    }
    catch (err) {
        return { error: `Error reading file: ${err}` };
    }
}
function resolveContractGroupsPayload(contents) {
    var _a, _b, _c, _d, _e;
    const nsp = (_a = contents.namespace) === null || _a === void 0 ? void 0 : _a.trim();
    if (!nsp) {
        return { error: `No "namespace" detected.` };
    }
    const groups = (contents.groups || []).filter((v) => !!v);
    if (!groups.length) {
        return { error: `No contract groups found to sync.` };
    }
    const resolvedGroups = [];
    let i = 1;
    for (const group of groups) {
        const name = (_b = group.name) === null || _b === void 0 ? void 0 : _b.trim();
        if (!name) {
            return { error: `No "name" given for contract group ${i}.` };
        }
        if (!group.abi) {
            const defaultAbiLocation = [constants_1.default.DEFAULT_ABIS_FOLDER, `${name}.json`].join('/');
            if (!(0, file_1.fileExists)(path_1.default.resolve(defaultAbiLocation))) {
                return {
                    error: `No "abi" given and no file at default location "./${defaultAbiLocation}"`,
                };
            }
            group.abi = defaultAbiLocation;
        }
        const { abi, isValid: isAbiValid } = (0, abi_1.resolveAbi)(group.abi);
        if (!isAbiValid)
            return {};
        const contracts = [];
        let j = 1;
        for (const contract of (group.contracts || []).filter((v) => !!v)) {
            const chainId = (_d = (_c = contract.chainId) === null || _c === void 0 ? void 0 : _c.toString()) === null || _d === void 0 ? void 0 : _d.trim();
            if (!chainId) {
                return { error: `No "chainId" given with contract ${j} in group "${name}"` };
            }
            if (!chains_1.chainIdsSet.has(chainId)) {
                return { error: `Chain id not yet supported: ${chainId}` };
            }
            const address = (_e = contract.address) === null || _e === void 0 ? void 0 : _e.toLowerCase().trim();
            if (!address) {
                return { error: `No "address" given with contract ${j} in group "${name}"` };
            }
            if (!(0, validators_1.isValidAddress)(address)) {
                return {
                    error: `Invalid "address" given with contract ${j} in group "${name}": ${address}`,
                };
            }
            contracts.push({ chainId, address });
            j++;
        }
        resolvedGroups.push({
            name,
            abi,
            contracts,
            isFactoryGroup: group.isFactoryGroup === true,
        });
        i++;
    }
    const payload = { nsp, groups: resolvedGroups };
    return { payload };
}
function pollForRegistrationResult(job, sessionToken, openWhenDone) {
    return __awaiter(this, void 0, void 0, function* () {
        const progressBars = {};
        const diff = (0, ansi_diff_stream_1.default)();
        const defaultBarWidth = 45;
        const logProgress = (groups, spinnerTextLength, done) => {
            for (const group of groups) {
                for (const instance of group.instances || []) {
                    const { chainId, address } = instance;
                    const progressWidth = defaultBarWidth - chainId.toString().length;
                    const key = [chainId, address].join(':');
                    progressBars[key] =
                        progressBars[key] ||
                            (0, progress_string_1.default)({
                                width: progressWidth,
                                total: 1,
                                incomplete: ' ',
                                complete: '.',
                            });
                }
            }
            const lines = [];
            for (const group of groups) {
                lines.push('\n' + chalk_1.default.cyanBright(group.name + ':'));
                lines.push(chalk_1.default.gray(chalk_1.default.dim('—')));
                for (const instance of group.instances || []) {
                    const { chainId, address } = instance;
                    const key = [chainId, address].join(':');
                    const bar = progressBars[key];
                    const value = done ? 1 : instance.progress || 0;
                    const pct = Math.min(Math.ceil(value * 100), 100);
                    const formattedPct = pct === 100 ? chalk_1.default.cyanBright(`${pct}%`) : `${pct}%`;
                    lines.push(`${chalk_1.default.gray(chainId)}:${address} ${chalk_1.default.cyan(bar(value))} ${formattedPct}`);
                }
            }
            lines.push(' '.repeat(spinnerTextLength));
            diff.write(lines.join('\n'));
        };
        diff.pipe(process_1.default.stdout);
        const spinner = (0, ora_1.default)({ stream: process_1.default.stdout });
        let startedSpinner = false;
        let i = 0;
        while (true) {
            job = i === 0 ? job : yield client_1.client.getContractRegistrationJob(sessionToken, job.uid);
            i++;
            if (job.error && job.error.toLowerCase().includes('parsing json'))
                continue;
            if (job.failed || job.error) {
                spinner.stop();
                (0, logger_1.logFailure)(job.error || 'Syncing contracts failed.');
                return;
            }
            const groups = job.groups || [];
            const allInstances = groups.map((g) => g.instances || []).flat();
            const numInstances = allInstances.length;
            const contractsEnglish = numInstances === 1 ? '1 contract' : `${numInstances} contracts`;
            const spinnerText = `Syncing ${contractsEnglish}...`;
            const isComplete = job.status === types_1.ContractRegistrationJobStatus.Complete;
            logProgress(groups, spinnerText.length, isComplete);
            if (!startedSpinner) {
                startedSpinner = true;
                spinner.text = spinnerText;
                spinner.start();
            }
            if (isComplete) {
                spinner.stop();
                (0, logger_1.logSuccess)(`Contracts in-sync.`);
                openWhenDone && (yield (0, open_1.default)((0, formatters_1.toContractGroupsUrl)(job.nsp)));
                return;
            }
            yield (0, time_1.sleep)(POLL_INTERVAL);
        }
    });
}
exports.default = addSyncContractsCmd;
//# sourceMappingURL=contracts.js.map