"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const logger_1 = require("../logger");
const constants_1 = __importDefault(require("../constants"));
const deno_1 = require("../utils/deno");
const msg_1 = __importDefault(require("../utils/msg"));
const formatters_1 = require("../utils/formatters");
const db_1 = require("../db");
const global_1 = require("../config/global");
const chains_1 = require("../utils/chains");
const date_1 = require("../utils/date");
const CMD = 'test';
function addTestCommand(program) {
    program
        .command(CMD)
        .description('Test one or more Live Tables')
        .argument('name', 'Name of the Live Table to test')
        .option('--recent', 'Test on the previous 30 days of data')
        .option('--days <type>', 'Number of days to fetch test data for')
        .option('--from <type>', 'Start date of the date range to fetch test data for')
        .option('--from-block <type>', 'Start block of the block range to fetch test data for')
        .option('--to <type>', 'End date of the range to fetch test data for')
        .option('--to-block <type>', 'End block of the block range to fetch test data for')
        .option('--all-time', 'Test over the entire date-range of input data used by the Live Table(s)')
        .option('--chains <type>', 'Chain ids to fetch test data for')
        .option('--keep-data', 'Whether to keep your existing test data from the previous run')
        .option('--port <type>', 'Port to run the testing server on')
        .action(test);
}
/**
 * Test one or more Live Tables.
 */
function test(name, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const { options, isValid } = validateOptions(opts || {});
        if (!isValid)
            return;
        // Ensure Deno is installed.
        const isDenoInstalled = (0, deno_1.ensureDenoInstalled)();
        if (!isDenoInstalled) {
            (0, logger_1.logWarning)(msg_1.default.INSTALL_DENO);
            return;
        }
        // Ensure psql is installed.
        if (!(0, db_1.psqlInstalled)()) {
            (0, logger_1.logWarning)(msg_1.default.INSTALL_PSQL);
            return;
        }
        // Ensure the testing database exists locally.
        const { error: dbError } = (0, db_1.upsertDatabase)(constants_1.default.LIVE_OBJECT_TESTING_DB_NAME);
        if (dbError) {
            (0, logger_1.logFailure)(`Failed to upsert the live object testing database locally: ${dbError}`);
            return;
        }
        // Get current project id.
        const { data: projectId, error } = (0, global_1.getCurrentProjectId)();
        if (error) {
            (0, logger_1.logFailure)(error);
            return;
        }
        if (!projectId) {
            (0, logger_1.logWarning)(msg_1.default.CHOOSE_PROJECT_MESSAGE);
            return;
        }
        // Get current project credentials from global spec creds file.
        const { data: creds, error: credsError } = (0, global_1.getProjectCreds)(projectId);
        if (credsError) {
            (0, logger_1.logFailure)(`Error finding project credentials: ${credsError}`);
            return;
        }
        if (!(creds === null || creds === void 0 ? void 0 : creds.apiKey)) {
            (0, logger_1.logWarning)(msg_1.default.CHOOSE_PROJECT_MESSAGE);
            return;
        }
        (0, deno_1.testLiveObject)(name, options, creds.apiKey);
    });
}
function validateOptions(options) {
    let { recent, days, from, fromBlock, to, toBlock, chains, allTime, keepData } = options;
    const chainIds = (chains || '')
        .split(',')
        .map((id) => id.trim())
        .filter((v) => !!v);
    days = days ? (0, formatters_1.toNumber)(days) : null;
    fromBlock = fromBlock ? (0, formatters_1.toNumber)(fromBlock) : null;
    toBlock = toBlock ? (0, formatters_1.toNumber)(toBlock) : null;
    from = from ? (0, formatters_1.toDate)(from) : null;
    to = to ? (0, formatters_1.toDate)(to) : null;
    // Validate chain ids.
    const invalidChainIds = chainIds.filter((id) => !chains_1.chainIdsSet.has(id));
    if (invalidChainIds.length) {
        (0, logger_1.logWarning)(`Invalid chain ids: ${invalidChainIds.join(', ')}`);
        return { isValid: false };
    }
    // Ensure negative numbers aren't given.
    if (days !== null && days < 0) {
        (0, logger_1.logWarning)(`--days can't be negative`);
        return { isValid: false };
    }
    if (fromBlock !== null && fromBlock < 0) {
        (0, logger_1.logWarning)(`--from-block can't be negative`);
        return { isValid: false };
    }
    if (toBlock !== null && toBlock < 0) {
        (0, logger_1.logWarning)(`--to-block can't be negative`);
        return { isValid: false };
    }
    // Prevent days and block ranges from meshing.
    if ((fromBlock || toBlock) && recent) {
        (0, logger_1.logWarning)(`--recent can't be used together with --from-block or --to-block`);
        return { isValid: false };
    }
    if (from && to && recent) {
        (0, logger_1.logWarning)(`--recent can't be used together with --from or --to`);
        return { isValid: false };
    }
    if ((fromBlock || toBlock) && days) {
        (0, logger_1.logWarning)(`--days can't be used together with --from-block or --to-block`);
        return { isValid: false };
    }
    if ((fromBlock || toBlock) && (from || to)) {
        (0, logger_1.logWarning)(`Can't blend blocks and dates when specifying a range.`);
        return { isValid: false };
    }
    if (from && to && days) {
        (0, logger_1.logWarning)(`--days can't be used when both --from and --to are also specified`);
        return { isValid: false };
    }
    if ((fromBlock || toBlock) && (from || to) && allTime) {
        (0, logger_1.logWarning)(`Can't specify a range when using --all-time`);
        return { isValid: false };
    }
    // Only 1 chain can be specified when using block ranges.
    if ((fromBlock || toBlock) && chainIds.length > 1) {
        (0, logger_1.logWarning)(`Can only use --from-block and --to-block with a single chain.`);
        return { isValid: false };
    }
    // Ensure ranges only move forwards in time or in series.
    if (fromBlock && toBlock && fromBlock > toBlock) {
        (0, logger_1.logWarning)(`Invalid block range: ${fromBlock} -> ${toBlock}`);
        return { isValid: false };
    }
    if (from && to && from > to) {
        (0, logger_1.logWarning)(`Invalid date range: ${from} -> ${to}`);
        return { isValid: false };
    }
    // Block ranges take priority over date ranges.
    if (fromBlock && from) {
        from = null;
    }
    if (toBlock && to) {
        to = null;
    }
    // Complete date range with "days" if given.
    if (from && days) {
        to = (0, date_1.addDays)(from, days);
    }
    if (to && days) {
        from = (0, date_1.subtractDays)(to, days);
    }
    if (!from && !to && days) {
        from = (0, date_1.subtractDays)(new Date(new Date().toUTCString()), days);
    }
    const opts = {
        recent,
        from,
        fromBlock,
        to,
        toBlock,
        chains: chainIds.join(','),
        allTime: !!allTime,
        keepData: !!keepData,
        port: (0, formatters_1.toNumber)(options.port) || constants_1.default.LIVE_OBJECT_TESTING_API_PORT,
    };
    return { options: opts, isValid: true };
}
exports.default = addTestCommand;
//# sourceMappingURL=test.js.map