"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateCreateTableMigrationFromLov = exports.getCurrentMigrations = exports.runMigration = exports.syncMigrations = exports.saveMigration = exports.splitMigrationName = exports.newMigration = exports.schemaName = void 0;
const file_1 = require("../utils/file");
const db_1 = require("../db");
const constants_1 = __importDefault(require("../constants"));
const path_1 = __importDefault(require("path"));
const chalk_1 = __importDefault(require("chalk"));
const logger_1 = require("../logger");
const formatters_1 = require("../utils/formatters");
const propertyTypes_1 = require("../utils/propertyTypes");
const pg_format_1 = require("pg-format");
const short_uuid_1 = __importDefault(require("short-uuid"));
const newConstraintName = (prefix) => `${prefix}_${short_uuid_1.default.generate().toLowerCase()}`;
const typeIdent = (type) => {
    return type.endsWith('[]') ? `${(0, pg_format_1.ident)(type.slice(0, -2))}[]` : (0, pg_format_1.ident)(type);
};
exports.schemaName = 'public';
const UP = 'up.sql';
const DOWN = 'down.sql';
const newVersion = () => new Date(new Date().toUTCString()).getTime();
function newMigration(action) {
    const version = newVersion();
    const name = [version, action].join('_');
    return { name, version };
}
exports.newMigration = newMigration;
function splitMigrationName(name) {
    const underscoreIndex = name.indexOf('_');
    if (!underscoreIndex)
        return null;
    let version = name.slice(0, underscoreIndex);
    const action = name.slice(version.length + 1);
    version = parseInt(version);
    if (Number.isNaN(version))
        return null;
    return [version, action];
}
exports.splitMigrationName = splitMigrationName;
function saveMigration(projectPath, migration, up = '', down = '') {
    const migrationsDirPath = path_1.default.join(projectPath, constants_1.default.SPEC_CONFIG_DIR_NAME, constants_1.default.MIGRATIONS_DIR_NAME);
    try {
        // Upsert migrations directory.
        (0, file_1.fileExists)(migrationsDirPath) || (0, file_1.createDir)(migrationsDirPath);
        // Create versioned migration directory.
        const versionDir = path_1.default.join(migrationsDirPath, migration.name);
        (0, file_1.createDir)(versionDir);
        // Create empty up/down migration files for this version.
        (0, file_1.createFileWithContents)(path_1.default.join(versionDir, UP), up);
        (0, file_1.createFileWithContents)(path_1.default.join(versionDir, DOWN), down);
        return { versionDir };
    }
    catch (error) {
        return { error };
    }
}
exports.saveMigration = saveMigration;
function syncMigrations(migrationsDir, url, env, logWhenNoAction = true) {
    return __awaiter(this, void 0, void 0, function* () {
        // Get list of current migrations.
        const { data: currentMigrations, error } = getCurrentMigrations(migrationsDir);
        if (error) {
            (0, logger_1.logFailure)(`Error while inspecting current migrations: ${error}`);
            return false;
        }
        if (!currentMigrations.length) {
            logWhenNoAction && (0, logger_1.log)(`Migrations up-to-date.`);
            return true;
        }
        // Get the latest migration version listed in the database.
        const { data: latestDbVersion, error: queryError } = (0, db_1.getLatestMigrationVersion)(url);
        if (queryError) {
            (0, logger_1.logFailure)(`Error checking database for latest migration version: ${queryError}`);
            return false;
        }
        // Find the migrations that haven't run yet.
        const migrationsToRun = currentMigrations.filter(({ version }) => version > latestDbVersion);
        if (!migrationsToRun.length) {
            logWhenNoAction && (0, logger_1.log)(`Migrations up-to-date.`);
            return true;
        }
        (0, logger_1.logSuccess)(`${migrationsToRun.length} new migrations detected`);
        (0, logger_1.log)(chalk_1.default.gray(`- - - - - - - - - - - - -`));
        (0, logger_1.log)(migrationsToRun
            .map((m, i) => {
            const [_, action] = splitMigrationName(m.name);
            return `${chalk_1.default.gray(`${i + 1}.`)} ${action}`;
        })
            .join('\n'));
        (0, logger_1.log)(chalk_1.default.gray(`- - - - - - - - - - - - -`));
        for (let i = 0; i < migrationsToRun.length; i++) {
            const migration = migrationsToRun[i];
            const success = yield runMigration(url, migrationsDir, migration, i + 1);
            if (!success)
                return false;
        }
        (0, logger_1.log)(chalk_1.default.greenBright(`Successfully ran ${migrationsToRun.length} migrations.`));
        return true;
    });
}
exports.syncMigrations = syncMigrations;
function runMigration(url, migrationsDir, migration, number, direction = UP) {
    return __awaiter(this, void 0, void 0, function* () {
        if (![UP, DOWN].includes(direction)) {
            (0, logger_1.logWarning)(`Invalid migration direction: ${direction}.`);
            return false;
        }
        const { version, name } = migration;
        const [_, action] = splitMigrationName(name);
        (0, logger_1.log)(`\n${chalk_1.default.cyanBright(`${number}. ${action}`)}\n`);
        const filePath = path_1.default.join(migrationsDir, name, direction);
        if (!(0, file_1.fileExists)(filePath)) {
            (0, logger_1.logFailure)(`No migration file found at: ${filePath}.`);
            return false;
        }
        if (!(yield (0, db_1.runSQLMigrationFileAgainstDB)(filePath, url))) {
            (0, logger_1.logFailure)('Migration failed. Stopping.');
            return false;
        }
        const { data: updateSuccess, error } = (0, db_1.updateLatestMigrationVersion)(url, version.toString());
        if (error) {
            (0, logger_1.logFailure)(error);
            return;
        }
        if (!updateSuccess) {
            (0, logger_1.logFailure)(`Failed to increment migration version to ${version} in the DB.`);
            return false;
        }
        return true;
    });
}
exports.runMigration = runMigration;
function getCurrentMigrations(migrationsDir) {
    let allEntries = [];
    try {
        allEntries = (0, file_1.getContentsOfFolder)(migrationsDir);
    }
    catch (error) {
        return { error };
    }
    if (!allEntries.length)
        return [];
    const migrations = [];
    for (const name of allEntries) {
        const [version, action] = splitMigrationName(name) || [];
        if (!version || !action)
            continue;
        let hasMigrationStructure = false;
        try {
            const migrationPath = path_1.default.join(migrationsDir, name);
            hasMigrationStructure =
                (0, file_1.isDir)(migrationPath) &&
                    (0, file_1.fileExists)(path_1.default.join(migrationPath, UP)) &&
                    (0, file_1.fileExists)(path_1.default.join(migrationPath, UP));
        }
        catch (err) {
            hasMigrationStructure = false;
        }
        if (!hasMigrationStructure)
            continue;
        migrations.push({ name, version });
    }
    return { data: migrations.sort((a, b) => a.version - b.version) };
}
exports.getCurrentMigrations = getCurrentMigrations;
function generateCreateTableMigrationFromLov(lov, tableName) {
    const columns = lov.properties.map(({ name, type }) => ({
        name: (0, formatters_1.camelToSnake)(name),
        type: (0, propertyTypes_1.guessColType)(type),
    }));
    const pkColumnNames = lov.uniqueBy.map(formatters_1.camelToSnake);
    // Create new table and add add primary keys.
    const [createTableSql, dropTableSql] = buildTableSql(tableName, columns);
    const addPrimaryKeySql = buildPrimaryKeySql(tableName, pkColumnNames);
    // Index by block_number/chain_id, as well as the primary timestamp.
    const indexes = [
        ['block_number', 'chain_id'],
        [(0, formatters_1.camelToSnake)(lov.primaryTimestampProperty)],
    ].map((columnNames) => buildIndexSql(tableName, columnNames));
    const upIndexes = [];
    let downIndexes = [];
    for (const [addIndex, dropIndex] of indexes) {
        upIndexes.push(addIndex);
        downIndexes.push(dropIndex);
    }
    downIndexes = downIndexes.reverse();
    const up = [createTableSql, addPrimaryKeySql, ...upIndexes];
    const down = [...downIndexes, dropTableSql];
    const migration = newMigration(`create_${tableName}`);
    return {
        up: sqlStatementsAsTx(up),
        down: sqlStatementsAsTx(down),
        migration,
    };
}
exports.generateCreateTableMigrationFromLov = generateCreateTableMigrationFromLov;
function buildTableSql(tableName, columns) {
    const columnStatements = columns.map((c) => buildColumnSql(c));
    return [
        `create table ${(0, pg_format_1.ident)(exports.schemaName)}.${(0, pg_format_1.ident)(tableName)} (${columnStatements.join(', ')})`,
        `drop table ${(0, pg_format_1.ident)(exports.schemaName)}.${(0, pg_format_1.ident)(tableName)}`,
    ];
}
function buildColumnSql(column) {
    return [(0, pg_format_1.ident)(column.name), typeIdent(column.type)].join(' ');
}
function buildPrimaryKeySql(tableName, columnNames) {
    const constraintName = newConstraintName('pk');
    return [
        `alter table ${(0, pg_format_1.ident)(exports.schemaName)}.${(0, pg_format_1.ident)(tableName)}`,
        `add constraint ${(0, pg_format_1.ident)(constraintName)}`,
        `primary key (${columnNames.map(pg_format_1.ident).join(', ')})`,
    ].join(' ');
}
function buildIndexSql(tableName, columnNames) {
    const indexName = newConstraintName('idx');
    return [
        `create index ${(0, pg_format_1.ident)(indexName)} on ${(0, pg_format_1.ident)(exports.schemaName)}.${(0, pg_format_1.ident)(tableName)} (${columnNames
            .map(pg_format_1.ident)
            .join(', ')})`,
        `drop index ${(0, pg_format_1.ident)(exports.schemaName)}.${(0, pg_format_1.ident)(indexName)}`,
    ];
}
function sqlStatementsAsTx(statements) {
    return ['BEGIN;', ...statements.map((s) => `    ${s};`), 'COMMIT;'].join('\n');
}
//# sourceMappingURL=migrations.js.map