"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.clearEventCursors = exports.resolveDBConnectionParams = exports.psqlInstalled = exports.runSQLMigrationFileAgainstDB = exports.updateLatestMigrationVersion = exports.getLatestMigrationVersion = exports.specSchemaTablesExist = exports.upsertDatabase = exports.getCurrentDbUser = exports.specUserExists = exports.upsertSpecUser = exports.initDatabase = void 0;
const node_child_process_1 = require("node:child_process");
const logger_1 = require("../logger");
const path_1 = __importDefault(require("path"));
const pw_1 = require("../utils/pw");
const constants_1 = __importDefault(require("../constants"));
const pg_format_1 = require("pg-format");
function initDatabase(dbName, url) {
    return __awaiter(this, void 0, void 0, function* () {
        // Upsert the "spec" database user.
        const { newlyAssignedPassword, error: upsertError } = upsertSpecUser(url);
        if (upsertError)
            return { error: upsertError };
        // Ensure db hasn't already been initialized.
        const { data: alreadyInitializedDb, error: schemaError } = specSchemaTablesExist(dbName, url);
        if (schemaError)
            return { error: schemaError };
        if (alreadyInitializedDb)
            return { newlyAssignedPassword, alreadyInitialized: true, error: null };
        const initFilePath = path_1.default.resolve(__dirname, '..', 'files', 'init.sql');
        (0, logger_1.log)('Initializing database for spec...');
        // Apply init script to db.
        const conn = url ? url : `-d ${dbName}`;
        try {
            (0, node_child_process_1.execSync)(`psql ${conn} -f ${initFilePath}`);
        }
        catch (error) {
            return { error };
        }
        return { newlyAssignedPassword, error: null };
    });
}
exports.initDatabase = initDatabase;
function upsertSpecUser(url) {
    const { data: exists, error } = specUserExists(url);
    if (error)
        return { error };
    if (exists)
        return { error: null };
    (0, logger_1.log)('Creating spec user...');
    let cmd = 'psql';
    let password = null;
    if (url) {
        cmd += ` ${url}`;
        password = url.includes('localhost') ? null : (0, pw_1.newPassword)();
    }
    const pwClause = password ? ` with password '${password}'` : '';
    cmd += ` -c "create user spec${pwClause}"`;
    try {
        (0, node_child_process_1.execSync)(cmd, { stdio: 'pipe' });
    }
    catch (err) {
        return { error: err };
    }
    if (password) {
        (0, logger_1.logSuccess)(`Created new database user:\nRole: "spec"\nPassword: "${password}"`);
    }
    return { newlyAssignedPassword: password, error: null };
}
exports.upsertSpecUser = upsertSpecUser;
function specUserExists(url) {
    let specUserExists = false;
    const cmd = url ? `psql ${url}` : 'psql';
    try {
        const out = (0, node_child_process_1.execSync)(`${cmd} -c "select usename from pg_catalog.pg_user where usename = 'spec'"`);
        const table = out.toString().trim();
        specUserExists = table.split('\n').length > 3;
    }
    catch (error) {
        return { error };
    }
    return { data: specUserExists };
}
exports.specUserExists = specUserExists;
function getCurrentDbUser() {
    try {
        const out = (0, node_child_process_1.execSync)(`psql -c "select current_user"`);
        return { data: (out.toString().trim().split('\n')[2] || '').trim() };
    }
    catch (error) {
        return { error };
    }
}
exports.getCurrentDbUser = getCurrentDbUser;
function upsertDatabase(name) {
    let exists = false;
    try {
        const out = (0, node_child_process_1.execSync)(`psql -c "select datname FROM pg_catalog.pg_database where lower(datname) = lower('${name}')"`);
        exists = out.toString().trim().split('\n').length > 3;
    }
    catch (err) {
        return { error: err };
    }
    if (exists) {
        return { error: null };
    }
    try {
        (0, node_child_process_1.execSync)(`createdb ${name}`);
    }
    catch (err) {
        return { error: err };
    }
    return { error: null };
}
exports.upsertDatabase = upsertDatabase;
function specSchemaTablesExist(dbName, url) {
    let specTablesExist = false;
    const conn = url ? url : `-d ${dbName}`;
    try {
        const out = (0, node_child_process_1.execSync)(`psql ${conn} -c "select tablename from pg_tables where schemaname = 'spec'"`);
        const table = out.toString().trim();
        specTablesExist = table.split('\n').length > 3;
    }
    catch (error) {
        return { error };
    }
    return { data: specTablesExist };
}
exports.specSchemaTablesExist = specSchemaTablesExist;
function getLatestMigrationVersion(url) {
    try {
        const out = (0, node_child_process_1.execSync)(`psql ${url} -c "select version from spec.migrations"`);
        const lines = out.toString().trim().split('\n');
        if (lines.length <= 3) {
            return { data: 0, rowExists: false };
        }
        const latestVersion = lines
            .slice(2, lines.length - 1)
            .map((v) => parseInt(v))
            .filter((v) => !Number.isNaN(v))
            .sort((a, b) => b - a)[0] || 0;
        return { data: latestVersion, rowExists: true };
    }
    catch (error) {
        return { error };
    }
}
exports.getLatestMigrationVersion = getLatestMigrationVersion;
function updateLatestMigrationVersion(url, version) {
    try {
        const { rowExists, error } = getLatestMigrationVersion(url);
        if (error)
            return { error };
        const cmd = rowExists
            ? `update spec.migrations set version = ${(0, pg_format_1.literal)(version)}`
            : `insert into spec.migrations (version) values (${(0, pg_format_1.literal)(version)}) on conflict (version) do nothing`;
        (0, node_child_process_1.execSync)(`psql ${url} -c "${cmd}"`);
        const { data: updatedVersion, error: checkError } = getLatestMigrationVersion(url);
        if (checkError)
            return { error: checkError };
        return { data: updatedVersion === parseInt(version) };
    }
    catch (error) {
        return { error };
    }
}
exports.updateLatestMigrationVersion = updateLatestMigrationVersion;
function runSQLMigrationFileAgainstDB(filePath, url) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((res, _) => __awaiter(this, void 0, void 0, function* () {
            try {
                const child = (0, node_child_process_1.spawn)('psql', [url, '-f', filePath]);
                child.stdout.setEncoding('utf8');
                child.stdout.on('data', function (data) {
                    process.stdout.write(data.toString());
                });
                let success = true;
                child.stderr.setEncoding('utf8');
                child.stderr.on('data', function (data) {
                    success = false;
                    process.stdout.write(data.toString());
                });
                child.on('close', () => {
                    res(success);
                });
            }
            catch (error) {
                console.log(error);
                res(false);
            }
        }));
    });
}
exports.runSQLMigrationFileAgainstDB = runSQLMigrationFileAgainstDB;
function psqlInstalled() {
    try {
        const out = (0, node_child_process_1.execSync)('which psql');
        const path = out.toString().trim();
        return !!path.length;
    }
    catch (error) {
        return false;
    }
}
exports.psqlInstalled = psqlInstalled;
function resolveDBConnectionParams(dbConfig) {
    let dbUser = dbConfig.user;
    if (!dbUser) {
        const { data: user, error } = getCurrentDbUser();
        if (error)
            return { error };
        dbUser = user;
    }
    if (!dbUser) {
        return { error: `Database user not specified.` };
    }
    if (!dbConfig.name) {
        return { error: `Database name not specified.` };
    }
    const connParams = {
        user: dbUser,
        password: dbConfig.password,
        host: dbConfig.host || 'localhost',
        port: Number(dbConfig.port || constants_1.default.DB_PORT),
        name: dbConfig.name,
    };
    return { data: connParams };
}
exports.resolveDBConnectionParams = resolveDBConnectionParams;
function clearEventCursors(url) {
    try {
        (0, node_child_process_1.execSync)(`psql ${url} -c "delete from spec.event_cursors"`);
    }
    catch (err) {
        (0, logger_1.logWarning)(`Failed to clear event cursors: ${JSON.stringify(err)}`);
    }
}
exports.clearEventCursors = clearEventCursors;
//# sourceMappingURL=index.js.map