"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.upsertLiveObject = exports.upsertDenoImports = exports.upsertDenoConfig = exports.upsertVsCodeSettings = exports.upsertGitignore = exports.upsertReadme = void 0;
const file_1 = require("../utils/file");
const path_1 = __importDefault(require("path"));
const logger_1 = require("../logger");
const formatters_1 = require("../utils/formatters");
const DEFAULT_VERSION = '0.0.1';
const dirs = {
    VSCODE: '.vscode',
};
const files = {
    GITIGNORE: '.gitignore',
    VSCODE_SETTINGS: 'settings.json',
    DENO_CONFIG: 'deno.json',
    DENO_IMPORTS: 'imports.json',
    README: 'README.md',
    SPEC: 'spec.ts',
    MANIFEST: 'manifest.json',
};
const gitIgnoreContents = `node_modules/
**/node_modules/
*.env
deno.lock`;
const vsCodeSettingsContents = `{
    "deno.enable": true,
    "deno.importMap": "./imports.json",
    "editor.defaultFormatter": "denoland.vscode-deno"
}`;
const denoConfigContents = `{
    "compilerOptions": {
        "strictPropertyInitialization": false,
        "strict": false
    }
}`;
const denoImportsContents = `{
    "imports": {
        "@spec.dev/core": "https://esm.sh/@spec.dev/core@0.0.142"
    }
}`;
const readmeContents = `TODO`;
const manifestFileContents = (namespace, name, version, displayName, description) => JSON.stringify({
    namespace,
    name,
    version,
    displayName: displayName || '',
    description: description || '',
}, null, 4);
const specFileContents = (name, description) => {
    const imports = ['Spec', 'LiveTable', 'Property', 'Event', 'OnEvent', 'Address'];
    let contents = `import { ${imports.join(', ')} } from '@spec.dev/core'

/**
 * ${description || 'TODO: ...'}
 */
@Spec({ 
    uniqueBy: ['someProperty', 'chainId'] 
})
class ${name} extends LiveTable {

    @Property()
    someProperty: Address

    // ==== Event Handlers ===================
    
    @OnEvent('namespace.ContractName.EventName')
    onSomeEvent(event: Event) {
        this.someProperty = event.data.someProperty
    }
}

export default ${name}`;
    return contents;
};
const fileContents = {
    GITIGNORE: gitIgnoreContents,
    VSCODE_SETTINGS: vsCodeSettingsContents,
    DENO_CONFIG: denoConfigContents,
    DENO_IMPORTS: denoImportsContents,
    README: readmeContents,
    manifest: manifestFileContents,
    spec: specFileContents,
};
function upsertFile(cwd, id) {
    const filePath = path_1.default.join(cwd, files[id]);
    (0, file_1.fileExists)(filePath) || (0, file_1.createFileWithContents)(filePath, fileContents[id]);
}
function upsertReadme(cwd) {
    upsertFile(cwd, 'README');
}
exports.upsertReadme = upsertReadme;
function upsertGitignore(cwd) {
    upsertFile(cwd, 'GITIGNORE');
}
exports.upsertGitignore = upsertGitignore;
function upsertVsCodeSettings(cwd) {
    const vsCodeFolderPath = path_1.default.join(cwd, dirs.VSCODE);
    const vsCodeSettingsPath = path_1.default.join(vsCodeFolderPath, files.VSCODE_SETTINGS);
    (0, file_1.fileExists)(vsCodeFolderPath) || (0, file_1.createDir)(vsCodeFolderPath);
    (0, file_1.fileExists)(vsCodeSettingsPath) ||
        (0, file_1.createFileWithContents)(vsCodeSettingsPath, fileContents.VSCODE_SETTINGS);
}
exports.upsertVsCodeSettings = upsertVsCodeSettings;
function upsertDenoConfig(cwd) {
    upsertFile(cwd, 'DENO_CONFIG');
}
exports.upsertDenoConfig = upsertDenoConfig;
function upsertDenoImports(cwd) {
    upsertFile(cwd, 'DENO_IMPORTS');
}
exports.upsertDenoImports = upsertDenoImports;
function upsertLiveObject(cwd, namespace, name, displayName, description) {
    const liveObjectId = (0, formatters_1.toNamespacedVersion)(namespace, name, DEFAULT_VERSION);
    const liveObjectFolderPath = path_1.default.join(cwd, name);
    // Ensure Live Object doesn't already exist.
    if ((0, file_1.fileExists)(liveObjectFolderPath)) {
        (0, logger_1.logWarning)(`Live Object already exists in folder "./${name}"`);
        return { liveObjectId, success: false };
    }
    // Create folder with the Live Object's name.
    (0, file_1.createDir)(liveObjectFolderPath);
    // Create "manifest.json" file.
    (0, file_1.createFileWithContents)(path_1.default.join(liveObjectFolderPath, files.MANIFEST), fileContents.manifest(namespace, name, DEFAULT_VERSION, displayName, description));
    // Create "spec.ts" file.
    (0, file_1.createFileWithContents)(path_1.default.join(liveObjectFolderPath, files.SPEC), fileContents.spec(name, description));
    return { liveObjectId, success: true };
}
exports.upsertLiveObject = upsertLiveObject;
//# sourceMappingURL=liveObjectTemplate.js.map